{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Some simple applications of top-down discrimination
    Copyright   : Fritz Henglein, 2007
    License     : All Rights Reserved

    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module DiscApps where

import Order
import SDisc
import Equiv
import Disc
import Char (toUpper)

-- Word occurrences

-- Word occurrences, alphabetically sorted
occsO :: [(String, Int)] → [[Int]]
occsO = sdisc ordString8

-- Word occurrences, reverse order of occurrence in input
occsE :: [(String, Int)] → [[Int]]
occsE = disc eqString8

-- Alphabetic order, case-insensitive
ordString8Ins :: Order String 
ordString8Ins = listL (MapO toUpper ordChar8)

-- Word occurrences, case-insensitive, alphabetic order
occsCaseInsO :: [(String, Int)] → [[Int]] 
occsCaseInsO = sdisc ordString8Ins

-- Word occurrences, case-insensitive, reverse order of occurrence in input
occsCaseInsE :: [(String, Int)] → [[Int]] 
occsCaseInsE = disc (Equiv ordString8Ins)

-- Test of occurrence functions
e1 = map (\occs -> occs [("I", 1), ("do", 3), ("believe", 6), ("yes", 14), ("I", 18), ("DO", 20)])
         [occsO, occsE, occsCaseInsO, occsCaseInsE]


-- Anagram classes 

-- Anagram classes, alphabetic order of word signature (reverse-alphabetically ordered)
anagramsO :: [String] -> [[String]]
anagramsO = spart (multiset ordChar8)

-- Anagram classes, order of occurrence
anagramsE :: [String] → [[String]] 
anagramsE = part (BagE eqChar8)

-- anagram classes, case insensitive, alphabetic order of signature
anagramsCaseInsO :: [String] -> [[String]]
anagramsCaseInsO = spart (multiset (MapO toUpper ordChar8))

-- anagram classes, case insensitive, alphabetic order of signature
anagramsCaseInsE :: [String] -> [[String]]
anagramsCaseInsE = part (BagE (MapE toUpper eqChar8))

-- equivalence for generalized anagrams, with sets as elements instead of characters
bsbOrder :: Equiv [[Int]]
bsbOrder = BagE (SetE eqNat8)

-- Test of anagram functions
e2 = map (\anagrams -> anagrams ["Dogs", "and", "only", "dogs", "believe", "in", "gods"])
         [anagramsO, anagramsE, anagramsCaseInsO, anagramsCaseInsE]


-- Lexicographic sorting

-- Lexicographic sorting of strings (lists of characters): ascending, descending, case-insensitive ascending
lexUp = dsort ordString8
lexDown = dsort (listL (Inv ordChar8))
lexUpCaseIns = dsort (listL (MapO toUpper ordChar8))

-- Lexicographic sorting of lists of strings: ascending, descending, case-insensitive ascending
lexUp2 = dsort (listL ordString8) 
lexDown2 = dsort (Inv (listL ordString8)) 
lexUpCaseIns2 = dsort (listL (listL (MapO toUpper ordChar8)))

-- Test of lexicographic sorting
e3 = map (\lex -> lex ["abc", "AbC", "zab", "YY"])
         [lexUp, lexDown, lexUpCaseIns]
e4 = map (\lex2 -> lex2 [["abc", "AbC", "zab", "YY"], ["bcd"], ["BCd", "BDE"]])
         [lexUp2, lexDown2, lexUpCaseIns2]


-- Type isomorphism

data TypeExp = TCons String [TypeExp] | 
               Prod TypeExp TypeExp 
     deriving Show

data TypeExp2 = TCons2 String [TypeExp2] 
              | Prod2 [TypeExp2]

unTypeExp2 (TCons2 v cts) = Left (v, cts)
unTypeExp2 (Prod2 cts) = Right cts

trans (Prod t1 t2) = Prod2 (traverse (Prod t1 t2) [])
trans (TCons c ts)  = TCons2 c (map trans ts)
traverse (Prod t1 t2) rem = traverse t1 (traverse t2 rem)
traverse (TCons c ts) rem = TCons2 c (map trans ts) : rem

prod2 :: Equiv TypeExp2
prod2 = MapE unTypeExp2 
        (SumE (PairE eqString8 (list prod2)) (list prod2))

typeIsoA :: [TypeExp] -> [[TypeExp]]
typeIsoA = part (MapE trans prod2)

prod3 :: Equiv TypeExp2
prod3 = MapE unTypeExp2 
        (SumE (PairE eqString8 (list prod2)) (BagE prod3))

typeIsoAC :: [TypeExp] -> [[TypeExp]]
typeIsoAC = part (MapE trans prod3)

-- Test of type isomorphism functions

intT = TCons "Int" []
boolT = TCons "Bool" []
listT t = TCons "List" [t]

e5 = map (\typeIso -> typeIso [ Prod (listT intT) (Prod boolT intT),
                                Prod (Prod (listT intT) boolT) intT,
                                Prod intT (Prod boolT (listT intT)) ])
         [typeIsoA, typeIsoAC]


-- Discriminatory joins

-- Sets, projections (functions), predicates (naively)
data Set a = Set [a] deriving Show
data Proj a b = Proj (a -> b)
data Pred a = Pred (a -> Bool) 

-- select, project, product (naively)
select :: Pred a -> Set a -> Set a
select (Pred c) (Set xs) = Set (filter c xs)

project :: Proj a b -> Set a -> Set b
project (Proj f) (Set xs) = Set (map f xs)

prod :: Set a -> Set b -> Set (a, b)
prod (Set xs) (Set ys) = Set [(x, y) | x <- xs, y <- ys]

-- equivalence join (naively)
joinN :: Proj a k -> Equiv k -> Proj b k -> Set a -> Set b -> Set (a, b)
joinN (Proj f1) e (Proj f2) s1 s2 =  
   select (Pred (\(x, y) -> eq e (f1 x) (f2 y))) (prod s1 s2) -- naive!

-- equivalence join, discriminatively
join :: Proj a k -> Equiv k -> Proj b k -> Set a -> Set b -> Set (a, b)
join (Proj f1) e (Proj f2) (Set xs) (Set ys) =
     Set [ (x, y) | (xs, ys) <- fprods, x <- xs, y <- ys ]
   where bs = disc e ([(f1 x, Left x) | x <- xs] ++ 
                      [(f2 y, Right y) | y <- ys])
         fprods = map split bs
         split :: [Either a b] -> ([a], [b])
         split [] = ([], [])
         split (v : vs) = let (lefts, rights) = split vs
                          in case v of { Left v' -> (v' : lefts, rights) ;
                                         Right v' -> (lefts, v' : rights) }
-- Test of discriminatory join
e6 = join (Proj fst) (Equiv evenOdd) (Proj fst)
          (Set [(5, "B"), (4, "A"), (7, "J")]) 
          (Set [(20, "P"), (88, "C"), (11, "E")])
