{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Equiv: exressions for denoting equivalence relations
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved
    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module Equiv where
import Order
import Foreign.Ptr
import Foreign.StablePtr
import System.IO.Unsafe

-- Equiv T: expresssions denoting equivalence relations over type T
data Equiv :: * -> *  where
     Equiv    :: Order t -> Equiv t
     SumE     :: Equiv t1 -> Equiv t2 -> Equiv (Either t1 t2)
     PairE    :: Equiv t1 -> Equiv t2 -> Equiv (t1, t2)
     MapE     :: (t1 -> t2) -> Equiv t2 -> Equiv t1
     BagE     :: Equiv t -> Equiv [t]
     SetE     :: Equiv t -> Equiv [t]

-- eq e : Characteristic function (equivalence/equality test) of equivalence relation denoted by e
eq :: Equiv t -> t -> t -> Bool
eq (Equiv r) x y = comp r x y == EQ
eq (SumE e1 e2) (Left x) (Left y) = eq e1 x y
eq (SumE e1 e2) (Right x) (Right y) = eq e2 x y
eq (SumE e1 e2) (Left _) (Right _) = False
eq (SumE e1 e2) (Right _) (Left _) = False
eq (PairE e1 e2) (x1, x2) (y1, y2) = eq e1 x1 y1 && eq e2 x2 y2
eq (MapE f e) x y = eq e (f x) (f y)
eq (BagE e) [] [] = True
eq (BagE e) [] (_ : _) = False
eq (BagE e) (x : xs') ys = 
   case delete e x ys of { Just ys' -> eq (BagE e) xs' ys' ;
                           Nothing -> False }
   where 
    delete :: Equiv t -> t -> [t] -> Maybe [t]
    delete e v = subtract' []
       where subtract' accum [] = Nothing
             subtract' accum (x : xs) = 
                if eq e x v then Just (accum ++ xs) else subtract' (x : accum) xs
eq (SetE e) xs ys = all (member e xs) ys && all (member e ys) xs
  where
    member :: Equiv t -> [t] -> t -> Bool
    member e [] v = False
    member e (x : xs) v = eq e v x || member e xs v

instance Show (Equiv t) where 
  show (Equiv r) = "(Equiv " ++ show r ++ ")"
  show (SumE e1 e2) = "(SumE " ++ show e1 ++ " " ++ show e2 ++ ")"
  show (PairE e1 e2) = "(PairE " ++ show e1 ++ " " ++ show e2 ++ ")"
  show (MapE f r) = "(MapE " ++ "<fcn> " ++ show r ++ ")"
  show (BagE e) = "(BagE " ++ show e ++ ")"
  show (SetE e) = "(BagE " ++ show e ++ ")"

-- Definable standard equivalence relations

-- Standard equality relations on primitive ordered types
eqUnit = Equiv ordUnit
eqBool = Equiv ordBool
eqNat8 = Equiv ordNat8
eqNat16 = Equiv ordNat16
eqInt32 = MapE splitW (PairE eqNat16 eqNat16)
   -- elides the subtraction of 2^31 in alternative definition eqInt32 = Equiv ordInt32
eqChar8 = Equiv ordChar8  
eqChar16 = Equiv ordChar16
eqString8 = Equiv ordString8

-- Equality relation on references (requires GHC library extension for breaking abstraction of references)
eqRef :: Eq v => Equiv v
eqRef = MapE ref2int eqInt32
  where ref2int :: Eq v => v -> Int
        ref2int x = unsafePerformIO (
                    do { sptr <- newStablePtr x ;
                         return (fromIntegral (ptrToIntPtr (castStablePtrToPtr sptr))) } )

-- list e: element-wise e-equivalence on lists
list :: Equiv t -> Equiv [t]
list e = MapE fromList (SumE (Equiv ordUnit) (PairE e (list e)))

-- andE e1 e2 : Greatest lower bound (intersection) of e1 and e2
--              x (andE e1 e2) y <=> x (e1) y and x (e2) y
andE :: Equiv t -> Equiv t -> Equiv t
andE e1 e2 = MapE dup (PairE e1 e2)
   where dup x = (x, x)

-- normalize : normalize an equivalence expression by pushing out Equiv constructors
normalize :: Equiv t -> Equiv t
normalize (Equiv (MapO f r)) = deforest (MapE f (normalize (Equiv r)))
normalize (Equiv (SumL r1 r2)) = SumE (normalize (Equiv r1)) (normalize (Equiv r2))
normalize (Equiv (PairL r1 r2)) = PairE (normalize (Equiv r1)) (normalize (Equiv r2))
normalize (Equiv (Inv r)) = normalize (Equiv r)
normalize (SumE e1 e2) = SumE (normalize e1) (normalize e2)
normalize (PairE e1 e2) = PairE (normalize e1) (normalize e2)
normalize (MapE f e) = deforest (MapE f (normalize e))
normalize (BagE e) = BagE (normalize e)
normalize (SetE e) = SetE (normalize e)
normalize e = e

-- deforest : fuse compositions of MapE applications
deforest (MapE f (MapE g e)) = MapE (g . f) e
deforest e = e
