{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Discriminator performance tests
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved

    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module Main where

import Order
import Equiv
import NatDisc -- basic multiset discriminator
import SDisc -- sorting discriminator
import Disc -- partitioning discriminator
import System.Random
import System.IO.Unsafe
import System.CPUTime
import System.Time
import CDisc -- discriminator using comparison-based sorting, sortBy, 
import QDisc -- Quicksort-based discriminator
import MDisc -- Mergesort-based discriminator
import Data.List (sort, sortBy, foldl')
import Char (ord)
import List (transpose)
import Control.Monad (liftM)

data Stream = S Int (IO Stream)

-- randList k m : generate a list of length m containing random integers from the range [0..k-1]
randList :: Int -> Int -> IO [Int]
randList k 0 = return []
randList k m = do { r <- getStdRandom (randomR (0, k-1)) ;
                    rs <- randList k (m-1) ;
                    return (r : rs) }

-- randLists k m n : generate a list of length n containing lists of length [0..m-1] with random integers in the range [0..k-1]
randLists :: Int -> Int -> Int -> IO [[Int]]
randLists k m n | n <= 0 = return []
randLists k m n = do { size <- getStdRandom (randomR (0, m)) ;
                       fst <- randList k size ;
                       rest <- randLists k m (n - size) ;
                       return (fst : rest) }

-- randLists2 k m n : generate a list of length n containing lists of length [0..m-1]^(1/i), i = 1, 2..., with random integers in [0..k-1]
randLists2 :: Int -> Float -> Int -> IO [[Int]]
randLists2 k p n | n <= 0 = return []
randLists2 k p n = do { size <- getStdRandom (randomR (1, round (p * fromIntegral n) `max` 1)) ;
                        fst <- randList k size ;
                        rest <- randLists2 k p (n - size) ;
                        return (fst : rest) }

-- size xs : size of xs, assuming fully-boxed representation of pairs and lists, size of pointer = 1, size of Int = 1
size :: [([Int], Int)] -> Int
size xs = 5 * length xs + foldr (+) 0 (map (size' . fst) xs) 
  where size' ys = 3 * length ys

chkSum :: [([Int], Int)] -> Int
chkSum xs = foldl' (+) 0 (concat (map fst xs))

minDiscPrefix :: Eq a => [a] -> [a] -> Int
minDiscPrefix [] ys = 0
minDiscPrefix xs [] = 0
minDiscPrefix (x : xs) (y : ys) = 
  if x == y then 1 + minDiscPrefix xs ys else 1

minDiscPrefixPairs :: Eq a => [[a]] -> [Int]
minDiscPrefixPairs [] = []
minDiscPrefixPairs [x] = []
minDiscPrefixPairs (x : xs@(y : ys)) = minDiscPrefix x y : minDiscPrefixPairs xs

minDiscPrefixes :: Ord a => [a] -> [a]
minDiscPrefixes [] = []
minDiscPrefixes [k] = [k]
minDiscPrefixes (k : ks@(l : ls)) = max k l : minDiscPrefixes ks

minDiscPrefixSum :: Eq a => [[a]] -> Int
minDiscPrefixSum xs = sum (minDiscPrefixes (0 : minDiscPrefixPairs xs))

minDiscPrefixSum2 :: Ord a => [([a], b)] -> Int
minDiscPrefixSum2 xs = minDiscPrefixSum (sort (map (sort . fst) xs))

fakeDisc :: [(k, v)] -> [[v]]
fakeDisc xs = [ [v] | (k, v) <- xs ]

discs = [ 
          ("fakeDisc", fakeDisc), 
          ("sortBy", cdisc (<=)), 
          ("qsort", qdisc (<=)),
          ("msort", mdisc (<=)), 
          ("sdisc", sdisc (listL ordNat16)), 
          ("disc", disc (list eqInt32)),
          ("sortBy (bag)", cdisc slte), 
          ("qsort (bag)", qdisc slte), 
          ("mdisc (bag)", mdisc slte), 
          ("sortBy (bag eff)", cdisc (<=) . map sortFst),
          ("qsort (bag eff)", qdisc (<=) . map sortFst), 
          ("msort (bag eff)", mdisc (<=) . map sortFst),
          ("sdisc (bag)", sdisc (multiset ordNat8)),
        ]
  where slte x y = sort x <= sort y
        scomp x y = compare (sort x) (sort y)
        sortFst (x, n) = (sort x, n)

ks = [256]
--ms = [0.3]
ms = [100]
ns = [100000, 200000, 300000, 400000, 500000, 600000, 700000, 800000, 900000, 1000000]

average :: [Integer] -> Integer
average xs = sum xs `div` (fromIntegral (length xs)) 

run k m n ds = do { 
                    ks <- randLists k m n;
                    let xs = zip ks ([1..] :: [Int])
                        run' d = do {
                                      startCPUTime <- getCPUTime ;
                                      res <- sum (map sum (d xs)) `seq` return () ;
                                      endCPUTime <- getCPUTime ;
                                      return ((endCPUTime - startCPUTime) `div` 1000000000) }
                        runIt = do {  
                                      cs <- return (chkSum xs);
                                      s <- return (size xs);
                                      ps <- return (minDiscPrefixSum2 xs);
                                      runTimes <- sequence (take 10 (repeat (mapM run' ds)));
                                      return (k, m, n, s, ps, map average (transpose runTimes))
                                    }

                    in runIt
                   }

toString :: (Int, Int, Int, Int, Int, [Integer]) -> String
toString (k, m, n, s, ps, runTimes) = 
     show k ++ ", " ++ show m ++ ", " ++ show n ++ ", " ++ show s ++ ", " ++ show ps ++
     concat (map (\t -> ", " ++ show t) runTimes) ++ "\n"

main = do { putStr ("k, m, n, size, minprefsize");
            mapM_ (\(dName, _) -> putStr (", " ++ dName)) discs;
            putStr ("\n");
            sequence_ [run k m n (map snd discs) >>= (putStr . toString) | k <- ks, m <- ms, n <- ns] }