{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Order: expressions denoting total preorders
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved
    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module Order where
import Data.Bits
import Char

-- Order T: expression denoting total preorders over type T
data Order :: * -> *  where
     Nat      ::  Int -> Order Int  
     Triv     ::  Order t      
     SumL     ::  Order t1 -> Order t2 -> Order (Either t1 t2)
     PairL    ::  Order t1 -> Order t2 -> Order (t1, t2)
     MapO     ::  (t1 -> t2) -> Order t2 -> Order t1 
     Inv      ::  Order t -> Order t 

-- lte r : characteristic function (inequality test) of ordering relation denoted by r
lte :: Order t -> t -> t -> Bool
lte (Nat n) x y = if 0 <= x && x <= n && 0 <= y && y <= n 
                  then x <= y 
                  else error "Outside of range" 
lte Triv _ _ = True
lte (SumL r1 r2) (Left x) (Left y) = lte r1 x y
lte (SumL r1 r2) (Left _) (Right _) = True
lte (SumL r1 r2) (Right _) (Left _) = False
lte (SumL r1 r2) (Right x) (Right y) = lte r2 x y
lte (PairL r1 r2) (x1, x2) (y1, y2) = lte r1 x1 y1 && if lte r1 y1 x1 then lte r2 x2 y2 else True
lte (MapO f r) x y = lte r (f x) (f y)
lte (Inv r) x y = lte r y x

-- comp r : comparison function for ordering relation denoted by r
comp :: Order t -> t -> t -> Ordering
comp (Nat n) x y = if 0 <= x && x <= n && 0 <= y && y <= n 
                   then compare x y
                   else error "Outside of range"
comp Triv _ _ = EQ
comp (SumL r1 r2) (Left x) (Left y) = comp r1 x y
comp (SumL r1 r2) (Left _) (Right _) = LT
comp (SumL r1 r2) (Right _) (Left _) = GT
comp (SumL r1 r2) (Right x) (Right y) = comp r2 x y
comp (PairL r1 r2) (x1, x2) (y1, y2) = 
   case comp r1 x1 y1 of { LT -> LT ;
                           EQ -> comp r2 x2 y2 ;
                           GT -> GT }
comp (MapO f r) x y = comp r (f x) (f y)
comp (Inv r) x y = comp r y x

instance Show (Order t) where
  show (Nat n) = "(Nat " ++ show n ++ ")"
  show Triv = "Triv"
  show (SumL r1 r2) = "(SumL " ++ show r1 ++ " " ++ show r2 ++ ")"
  show (PairL r1 r2) = "(PairL " ++ show r1 ++ " " ++ show r2 ++ ")"
  show (MapO f r) = "(MapO " ++ "<fcn> " ++ show r ++ ")"
  show (Inv r) = "(Inv " ++ show r ++ ")"

-- Definable standard ordering relations

-- ordUnit: standard order on unit type
ordUnit :: Order ()
ordUnit = Triv

-- bool: standard order on Bool: False < True
ordBool :: Order Bool
ordBool = MapO bool2sum (SumL ordUnit ordUnit)
  where bool2sum :: Bool -> Either () ()
        bool2sum False = Left ()
        bool2sum True = Right ()

-- ordNat8 : standard ascending order on Int, restricted to 8-bit nonnegative integers
ordNat8 :: Order Int
ordNat8 = Nat 255

-- ordNat16 : standard ascending order on Int, restricted to 16-bit nonnegative integers
ordNat16 :: Order Int
ordNat16 = Nat 65535

-- ordInt32 : standard order on 32-bit integers
-- Adds 10...0xB to number before splitting it into high- and low-order 16 bits
ordInt32 :: Order Int 
ordInt32 = MapO (splitW . (+ 2147483648)) (PairL ordNat16 ordNat16)

-- splitW x : split 32-bit integer into high and low 16 bits
splitW :: Int -> (Int, Int)
splitW x = (shiftR x 16 .&. 65535, x .&. 65535)

-- ordChar8 : standard alphabetic order on Char, restricted to 8-bit characters 
ordChar8 :: Order Char        
ordChar8 = MapO ord ordNat8

-- ordChar16 : standard alphabetic order on Char, restricted to 16-bit characters 
ordChar16 :: Order Char        
ordChar16 = MapO ord ordNat16

-- Definable order constructors

-- fromList: unfold-part of isomorphism between [t] and Either () (t, [t])
fromList :: [t] -> Either () (t, [t])
fromList []       = Left ()
fromList (x : xs) = Right (x, xs)

-- Lexicographic ordering on lists, with elements ordered according to r
listL :: Order t -> Order [t]
listL r = MapO fromList (SumL ordUnit (PairL r (listL r)))

-- ordString8 : standard alphabetic order on String, restricted to 8-bit characters
ordString8 :: Order String
ordString8 = listL ordChar8

-- Ordering of tagged values, but with Right elements first
sumR :: Order t1 -> Order t2 -> Order (Either t1 t2)
sumR r1 r2 = MapO flip (SumL r2 r1)
   where flip :: Either t1 t2 -> Either t2 t1
         flip (Left x) = Right x
         flip (Right y) = Left y

-- Lexicographic ordering on pairs, but with second component dominant
pairR :: Order t1 -> Order t2 -> Order (t1, t2)
pairR r1 r2 = MapO swap (PairL r2 r1)
    where swap :: (t1, t2) -> (t2, t1)
          swap (x, y) = (y, x)

-- Refinement of one ordering by another: 
-- x (Refine r1 r2) y <=> x (r1) y and (y (r1) x => x (r2) y)
refine :: Order t -> Order t -> Order t
refine r1 r2 = MapO dup (PairL r1 r2)
    where dup x = (x, x)

-- simplify: Algebraic simplification of order denotations
simplify :: Order t -> Order t
simplify (Inv Triv) = Triv
simplify r@(Inv (Nat _)) = r
simplify (Inv (SumL r1 r2)) = sumR (simplify (Inv r1)) (simplify (Inv r2))
simplify (Inv (PairL r1 r2)) = PairL (simplify (Inv r1)) (simplify (Inv r2))
simplify (Inv (MapO f r)) = MapO f (simplify (Inv r))
simplify (Inv (Inv r)) = simplify r

-- Additional examples of order denotations

-- evenOdd: even number < odd number, 
evenOdd :: Order Int 
evenOdd = MapO (`mod` 2) (Nat 2)

data Tree v = Node (v, [Tree v])
unNode (Node n) = n

-- tree: Standard order on trees
tree :: Order t -> Order (Tree t)
tree r = MapO unNode (PairL r (listL (tree r)))

-- treeK: Level-k cut-off order (subtrees at level >= k are considered equivalent)
treeK :: Int -> Order t -> Order (Tree t)
treeK 0 r = Triv
treeK k r = MapO unNode (PairL r (listL (treeK (k-1) r)))

-- flipFlop: orders lists by lexicographic ordering on last, first, next-to-last, second, etc., elements of lists
flipFlop :: Order t -> Order [t]
flipFlop r = 
  MapO (fromList . reverse) (SumL ordUnit (PairL r (flipFlop r)))
