{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : SDisc: sorting discriminators
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved

    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module SDisc where

import Order 
import Array (accumArray, elems)

-- Generic definition of reverse-stable sorting discriminators 
sdisc :: Order k -> [(k, v)] -> [[v]]
sdisc r []               = []
sdisc r [(k, v)]         = [[v]]
sdisc (Nat n) xs         = sdiscNat n xs
sdisc Triv xs            = [map snd xs]
sdisc (SumL r1 r2) xs    = sdisc r1 lefts ++ sdisc r2 rights
   where (lefts, rights) = split xs
sdisc (PairL r1 r2) xs   = [ vs | ys <- sdisc r1 [ (k1, (k2, v)) | ((k1, k2), v) <- xs ],
                                  vs <- sdisc r2 ys ]
sdisc (MapO f r) xs      = sdisc r [ (f k, v) | (k, v) <- xs ] 
sdisc (Inv r) xs         = reverse (sdisc r xs)

-- Bucket-sorting basic discriminator
sdiscNat :: Int -> [(Int, v)] -> [[v]]
sdiscNat n xs = filter (not . null) (bucket n update xs)
   where update vs v = v : vs

-- Bucket-sorting template routine
-- bucket n update [(k1, v1), ...(km, vm)]:
-- initialize new bucket table T[0..n-1];
-- for each i = 1..m, write update T[ki] vi into T[ki];
-- return [ T[i] | i <- [0..n] ]
bucket :: Int -> ([v] -> v -> [v]) -> [(Int, v)] -> [[v]]
bucket (n :: Int) update xs = elems (accumArray update [] (0, n) xs)

--Splitting tagged keys into left and right values
split :: [(Either k1 k2, v)] -> ([(k1, v)], [(k2, v)])
split = split' ([], [])
  where split' acc@(lefts, rights) [] = acc
        split' (lefts, rights) ((Left k, v) : xs) = split' ((k, v) : lefts, rights) xs
        split' (lefts, rights) ((Right k, w) : xs) = split' (lefts, (k, w) : rights) xs
      
-- Sorted partitioning from sorted discrimination
spart :: Order t -> [t] -> [[t]]
spart r xs = sdisc r [ (x, x) | x <- xs ]

-- Discriminative sorting from sorted partitioning
dsort :: Order t -> [t] -> [t]
dsort r xs = [ y | ys <- spart r xs, y <- ys ]

-- Unique sorting from ordered partitioning
usort :: Order t -> [t] -> [t]
usort r xs = [ head ys | ys <- spart r xs ]

-- Comparator defined from ordered discriminator
dcomp :: Order t -> t -> t -> Ordering
dcomp r x y = case sdisc r [(x, LT), (y, GT)] of {
                [_] -> EQ ;
                (o : _) : _ -> o }

-- More order expressions

-- Dershowitz-Manna multiset ordering on total preorders
multiset :: Order t -> Order [t]
multiset r = MapO (dsort (Inv r)) (listL r)

-- Dershowitz-Manna style set ordering on total preorders
set :: Order t -> Order [t]
set r = MapO (usort (Inv r)) (listL r)
